// Fill out your copyright notice in the Description page of Project Settings.

#pragma once

#include "CoreMinimal.h"
#include "FrameGrabber.h"
/**
 * 
 */
class FLiveFrameGrabber
{
public:
	/**
	 * Construct this frame grabber
	 *
	 * @param InViewport			The viewport we are to grab frames for
	 * @param DesiredBufferSize		The desired size of captured frames
	 * @param InPixelFormat			The desired pixel format to store captured frames as
	 * @param InNumSurfaces			The number of destination surfaces contained in our buffer
	 */
	FLiveFrameGrabber(FRenderTarget * RenderTarget,FIntPoint DesiredBufferSize, EPixelFormat InPixelFormat = PF_B8G8R8A8, uint32 NumSurfaces = 3);

	/** Destructor */
	~FLiveFrameGrabber();

public:

	/** Instruct the frame grabber to start capturing frames */
	void StartCapturingFrames();

	/** Check whether we're capturing frames or not */
	bool IsCapturingFrames() const;

	/** Instruct the frame grabber capture this frame, when it receives an event from slate */
	void CaptureThisFrame(FFramePayloadPtr Payload);

	/** Stop capturing frames */
	void StopCapturingFrames();

	/** Shut down this grabber, ensuring that any threaded operations are finished */
	void Shutdown();

public:

	/** Check whether we have any outstanding frames or not */
	bool HasOutstandingFrames() const;

	/** Retrieve any frames we may have captured */
	TArray<FCapturedFrameData> GetCapturedFrames();

protected:

	/** Callback for when a backbuffer is ready for reading (called on render thread) */
	void OnBackBufferReadyToPresentCallback(SWindow& SlateWindow, const FTexture2DRHIRef& BackBuffer);

	/** Called when the specified surface index has been locked for reading with the render target data (called on render thread)  */
	void OnFrameReady(int32 SurfaceIndex, FColor* ColorBuffer, int32 Width, int32 Height);

private:

	/** Non-copyable */
	FLiveFrameGrabber(const FLiveFrameGrabber&);
	FLiveFrameGrabber& operator=(const FLiveFrameGrabber&);

	/**
	 * Pointer to the window we want to capture.
	 * Only held for comparison inside OnBackBufferReadyToPresentCallback - never to be dereferenced or cast to an SWindow.
	 * Held as a raw pointer to ensure that no referenc counting occurs from the background thread in OnBackBufferReadyToPresentCallback.
	 */
	void* TargetWindowPtr;

	/** Delegate handle for the OnBackBufferReadyToPresent event */
	FDelegateHandle OnBackBufferReadyToPresent;

	/** Array of captured frames */
	TArray<FCapturedFrameData> CapturedFrames;

	/** Lock to protect the above array */
	mutable FCriticalSection CapturedFramesMutex;

	/** Array of surfaces that we resolve the viewport RHI to. Fixed allocation - should never be resized */
	struct FResolveSurface
	{
		FResolveSurface(EPixelFormat InPixelFormat, FIntPoint BufferSize) : Surface(InPixelFormat, BufferSize) {}

		FFramePayloadPtr Payload;
		FViewportSurfaceReader Surface;
	};
	TArray<FResolveSurface> Surfaces;

	/** Index into the above array to the next surface that we should use - only accessed on main thread */
	int32 CurrentFrameIndex;

	/** The index that we should capture the next rendered slate window into */
	int32 SlateRenderIndex;

	/** The total number of frames we are currently waiting on */
	FThreadSafeCounter OutstandingFrameCount;

	/** Only to be accessed from the render thread - array of frame payloads to be captured from the rendered slate window sorted first to last. */
	TArray<FFramePayloadPtr> RenderThread_PendingFramePayloads;

	int32 FrameGrabLatency;

	/** The current state of the grabber */
	enum class EFrameGrabberState
	{
		Inactive, Active, PendingShutdown
	};
	EFrameGrabberState State;

	/** The desired target size to resolve frames to */
	FIntPoint TargetSize;

	FRenderTarget* RenderTarget;
};
